use iron::headers;
use iron::middleware::BeforeMiddleware;
use iron::prelude::*;
use iron::typemap::Key;
use orizentic::{filedb, OrizenticCtx, Secret};
use params::{FromValue, Params};

use crate::cookies::{Cookie, CookieJar};

pub struct Authentication {
    pub auth: OrizenticCtx,
}

impl Key for Authentication {
    type Value = orizentic::VerifiedToken;
}

impl Authentication {
    pub fn new(secret: Secret, auth_db_path: String) -> Authentication {
        let claims = filedb::load_claims_from_file(&auth_db_path).expect("claims did not load");
        let orizentic = OrizenticCtx::new(secret, claims);
        Authentication { auth: orizentic }
    }

    fn authenticate_user(
        &self,
        token_str: String,
    ) -> Result<orizentic::VerifiedToken, orizentic::Error> {
        self.auth.decode_and_validate_text(token_str)
    }
}

impl BeforeMiddleware for Authentication {
    fn before(&self, req: &mut Request) -> IronResult<()> {
        let params = req.get_ref::<Params>().unwrap();
        let token = match params.get("auth").and_then(|v| String::from_value(v)) {
            Some(token_str) => self.authenticate_user(token_str).ok(),
            None => {
                let m_jar = req
                    .headers
                    .get::<headers::Cookie>()
                    .map(|cookies| CookieJar::from(cookies));
                m_jar
                    .and_then(|jar| jar.lookup("auth").cloned())
                    .and_then(|Cookie { value, .. }| self.authenticate_user(value.clone()).ok())
            }
        };
        token.map(|t| req.extensions.insert::<Authentication>(t));
        Ok(())
    }
}
